
volatile int timer_water_sensor;//таймер опроса датчиков воды(в секундах)
volatile int timer_weather_station;//таймер опроса датчиков погодной станции(в секундах)

//Буферные переменные для хранения текущих значений даты/времени(не работают вне цикла loop)
uint16_t rtc_year;
uint8_t rtc_month;
uint8_t rtc_day;
uint8_t rtc_hour;
uint8_t rtc_minute;
uint8_t rtc_second;
boolean errorr_rtc_state;
const char *errorr_rtc[]= {"",""};

//Буферные переменные для хранения текущих значений BM180(не работают вне цикла loop)
float temp_BM180;
float pres_BM180;
boolean errorr_BM180_state;
const char *errorr_BM180[]= {"",""};

//Буферные переменные для хранения текущих значений DHT21(не работают вне цикла loop)
float temp_DHT21;
int humi_DHT21;
boolean errorr_DHT21_state;
const char *errorr_DHT21[]= {"",""};

boolean prevstate_cws;// предыдущее состояние контакта датчика холодной воды
boolean prevstate_chs;// предыдущее состояние контакта датчика горячей воды

int imp;
int imp_period = 180;//период, измерения мощности в секундах
float imp_const = 0.625*(3600/imp_period);//константа для получения текущей мощности(умножается на колличество импульсов)
unsigned long cur_millis = 0;
unsigned long loop_millis = 0;

#define pin_chs 4//pin датчика горячей воды(chs - cold hot sensor)
#define pin_cws 3//pin датчика холодной воды(cws - cold water sensor)
#define pin_pem 2//pin импульсного выхода счетчика эллектроэнергии(pem - pulse electricity meter)
#define pin_dht21 5//pin погодного датчика DHT21

//Библиотека для работы с погодным датчиком BMP180
#include <Adafruit_BMP085.h>
Adafruit_BMP085 bmp;

//Библиотека для работы с часами DS1307
#include "RTClib.h"
RTC_DS1307 rtc;

//Библиотека для работы с датчиками серии DHT
#include <TroykaDHT.h>
DHT dht(pin_dht21, DHT21);


void setup(){
    Serial.begin(9600);
    
    pinMode(pin_cws, INPUT_PULLUP);//включаем резистор на данном пине
    digitalWrite(pin_cws, HIGH);//включаем резистор на данном пине(дополнительно, на всякий случай)

    if (! rtc.begin()) {
      errorr_rtc_state = true;
      errorr_rtc[0] = "1";
      errorr_rtc[1] = "Could not find RTC sensor";
      Serial.println("{\"error\":{\"subject\":\"RTC\",\"cod\":\"" + String(errorr_rtc[0]) + "\",\"text\":\"" + String(errorr_rtc[1])+ "\",\"date\":\"" + DateTimeNow() + "\"}}");
      while (1) {}
    }
    
    if (!bmp.begin()) {
      errorr_BM180_state = true;
      errorr_BM180[0] = "1";
      errorr_BM180[1] = "Could not find a valid BMP085 sensor, check wiring!";
      Serial.println("{\"error\":{\"subject\":\"BM180\",\"cod\":\"" + String(errorr_BM180[0]) + "\",\"text\":\"" + String(errorr_BM180[1])+ "\",\"date\":\"" + DateTimeNow() + "\"}}");
      while (1) {}
    }

    dht.begin();

    // инициализация Timer1
    cli();  // отключить глобальные прерывания
    TCCR1A = 0;   // установить регистр в 0
    TCCR1B = 0;   // установить регистр в 0
    OCR1A = 15624; // установка регистра совпадения (1 секунда)
    TCCR1B |= (1 << WGM12);  // включить CTC режим > сброс таймера по совпадению
    TCCR1B |= (1 << CS10);   // Установить биты CS10 CS12 на коэффициент деления 1024
    TCCR1B |= (1 << CS12);
    // F (ПРЕРЫВАНИЙ) = 16000000 / (1024 *(1 + 15624) = 1 s 
    TIMSK1 |= (1 << OCIE1A);  // включить прерывание по совпадению таймера 
    sei(); // включить глобальные прерывания

    pinMode(pin_pem, INPUT_PULLUP);
    attachInterrupt(0, energy_impulse, FALLING);
}

 
void loop(){
   if (((millis()/1000)-loop_millis) >= 2)
   {
//Работаем с временем, сохраним значения в буферные переменные, так как в явном виде, значения не работают вне цикла loop.
DateTime now = rtc.now();
rtc_year = now.year();
rtc_month = now.month();
rtc_day = now.day();
rtc_hour = now.hour();
rtc_minute = now.minute();
rtc_second = now.second();


temp_BM180 = bmp.readTemperature();
pres_BM180 = (bmp.readPressure()*0.007501);

dht.read();
      temp_DHT21 = dht.getTemperatureC();
      humi_DHT21 = dht.getHumidity();
/*
  switch(dht.getState()) {
    case DHT_OK:
      temp_DHT21 = dht.getTemperatureC();
      humi_DHT21 = dht.getHumidity();
     
    // ошибка контрольной суммы
    case DHT_ERROR_CHECKSUM:
      errorr_DHT21_state = true;
      errorr_DHT21[0] = "1";
      errorr_DHT21[1] = "Checksum error";

    // превышение времени ожидания
    case DHT_ERROR_TIMEOUT:
      errorr_DHT21_state = true;
      errorr_DHT21[0] = "2";
      errorr_DHT21[1] = "Time out error";
      
    // данных нет, датчик не реагирует или отсутствует
    case DHT_ERROR_NO_REPLY:
      errorr_DHT21_state = true;
      errorr_DHT21[0] = "3";
      errorr_DHT21[1] = "Sensor not connected";
  }
  */
  loop_millis = (millis()/1000);
   }
}

String energy_impulse() {
  /*
   Функция обработки импульса с счетчика эллектроэнергии.
   Каждый раз когда счетчик моргает светодиодом, получаем импульс, 1600 импульсов = 1 кв.ч
   1 минута = 43200 записей в месяц
   2 минуты = 21600 записей в месяц
   3 минуты = 14400 записей в месяц
   5 минут  = 8640  записей в месяц
   передаем данные в формате JSON.
   Пример:
   {
   "pulse":
    {
    "imp":"",
    "imp_period":"",
    "date":"2022:02:08 01:01:01"
    }
   }
  */
   imp++;
   if (((millis()/1000)-cur_millis) >= imp_period)
   {
    cur_millis = (millis()/1000);
   //Расчитаем потребление Ватт/ч за период (1000 импульсов/1600 импульсов * imp * 3600/imp_period)
   //,\"consumption\":\"" + String((imp*imp_const)) + "\" - Потребление в кв/ч
   Serial.flush();
   Serial.println("{\"pulse\":{\"imp\":\"" + String(imp) + "\",\"period\":\"" + String(imp_period) + "\",\"date\":\"" + DateTimeNow() + "\"}}");
   imp = 0;
   }
   
}

ISR(TIMER1_COMPA_vect)// Вызывается ISR(TIMER1_OVF_vect), это происходит всегда когда таймер переполняется
{
    timer_water_sensor++;
    timer_weather_station++;
    
    if(timer_water_sensor == 2)//интервал в секундах
    {
      timer_water_sensor = 0;
      water_sensor();
    }

    if(timer_weather_station == 300)//интервал в секундах
    {
      timer_weather_station = 0;
      weather_station();
    }
}

void water_sensor() {
  /*
   Функция обработки датчиков воды.
   Каждый раз когда на циферблате счетчика(в верхнем окошке) появляется 9, замыкается геркон.
   Раз в период, проверяем состояние контакта,
   если контакт был замкнут, а до этого разомкнут передаем данные в формате JSON.
   Пример:
   {
   "water_sensor":
    {
    "cws":"1",
    "chs":"1",
    "date":"2022:02:08 01:01:01"
    }
   }
   *cws - cold water sensor
   *chs - cold hot sensor
  */
    if(prevstate_cws == true && digitalRead(pin_cws) == false)
      {
        Serial.flush();
        Serial.println("{\"water_sensor\":{\"cws\":\"1\",\"date\":\"" + DateTimeNow() + "\"}}");
      }
    prevstate_cws = boolean(digitalRead(pin_cws));
}

void weather_station() {
    /*
   Функция обработки датчиков погодной станции.
   Каждый датчик обрабатывает и передает данные отдельно.
   - BM180 - датчик передает температуру в цельсиях(temp) и давление в мм.рт.ст(pressure)
   - DHT21 - датчик передает температуру в цельсиях(temp) и влажность в %(humidity)
   передаем данные в формате JSON. В случае ошибки возвращаем текст ошибки и код.
   Пример:
   {
   "BM180":
    {
    "temp":"",
    "pressure":"",
    "date":"2022:02:08 01:01:01"
    }
   }
  или в случае ошибки
   {
   "error":
    {
    "subject":"BM180",
    "cod":"1",
    "text":"Could not find a valid BMP085 sensor, check wiring!",
    "date":"2022:02:08 01:01:01"
    }
   }
  */
  if (!errorr_BM180_state)
  {
    Serial.flush();
    Serial.println("{\"BM180\":{\"temp\":\""+ String(temp_BM180) +"\",\"pressure\":\""+ String(pres_BM180) +"\",\"date\":\"" + DateTimeNow() + "\"}}");
  }
  else 
  {
    Serial.flush();
    Serial.println("{\"error\":{\"subject\":\"BM180\",\"cod\":\"" + String(errorr_BM180[0]) + "\",\"text\":\"" + String(errorr_BM180[1])+ "\",\"date\":\"" + DateTimeNow() + "\"}}");
  }

  if (!errorr_DHT21_state)
  {
    Serial.flush();
    Serial.println("{\"DHT21\":{\"temp\":\""+ String(temp_DHT21) +"\",\"humidity\":\""+ String(humi_DHT21) +"\",\"date\":\"" + DateTimeNow() + "\"}}");
  }
  else 
  {
    Serial.flush();
    Serial.println("{\"error\":{\"subject\":\"DHT21\",\"cod\":\"" + String(errorr_DHT21[0]) + "\",\"text\":\"" + String(errorr_DHT21[1])+ "\",\"date\":\"" + DateTimeNow() + "\"}}");
  }
}

String DateTimeNow() {
    /*
   Функция выводв текущей даты и времени в формате ISO 8601.
   Пример: 2022-2-9T13:27:25
  */
    return String(rtc_year) + "-" + String(rtc_month) + "-" + String(rtc_day) + "T" + String(rtc_hour) + ":" + String(rtc_minute) + ":" + String(rtc_second);
}
